@@ -206,7 +206,6 @@ class DynamoFL(_Base):
         self.wshost = self.host.replace('http', 'ws', 1)
         self.project_callbacks = {}
         self.on_round_threads = {}
-        # self.projects = {}
         self.project_participants = []
         self.datasources = {}
         self.instance_id = None
@@ -237,11 +236,6 @@ class DynamoFL(_Base):
             datasource_key = j['data']['datasourceKey']
             trainer_key = j['data']['trainerKey']
 
-            # self.projects[project_key] = {
-            #     'datasource_key': datasource_key,
-            #     'trainer_key': trainer_key
-            # }
-
             self.project_participants.append({
                 'project_key': project_key,
                 'datasource_key': datasource_key,
@@ -256,17 +250,7 @@ class DynamoFL(_Base):
         if 'data' in j and 'project' in j['data'] and 'key' in j['data']['project']:
             project_key = j['data']['project']['key']
         if j['event'] == 'project-complete':
-            for p in self.project_participants:
-                threads_key = get_threads_key(project_key, p['datasource_key'])
-                if threads_key in self.on_round_threads:
-                    self.on_round_threads[threads_key].kill()
-                if project_key == p['project_key']:
-                    self.on_round_threads[threads_key] = gevent.spawn(self.train_and_test_and_end_callback, p['datasource_key'], j['data']['project'])
-
-                    # del self.projects[project_key]
-            # if not len(self.project_participants):
-            #     self.ws.close()
-            #     return
+            self.project_participants = list(filter(lambda x : x['project_key'] != project_key, self.project_participants))
 
 
         elif j['event'] == 'round-complete':
@@ -291,6 +275,12 @@ class DynamoFL(_Base):
         print('Connection error:')
         print(error)
 
+    def _get_last_fed_model_round(self, current_round, is_complete):
+        if is_complete:
+            return current_round
+        else:
+            return current_round - 1
+
     def train_and_test_callback(self, datasource_key, project_info):
         project = _Project(self.token, self.host, project_info['key'], self.ws)
 
@@ -300,53 +290,49 @@ class DynamoFL(_Base):
             if project_info['key'] == p['project_key'] and datasource_key == p['datasource_key']:
                 trainer_key = p['trainer_key']
                 break
+
         train = self.datasources[datasource_key].trainers[trainer_key]['train']
         test = self.datasources[datasource_key].trainers[trainer_key]['test']
         model_path = 'models'
         if 'model_path' in self.datasources[datasource_key].trainers[trainer_key]:
             model_path = self.datasources[datasource_key].trainers[trainer_key]['model_path']
 
-        current_round = project_info['currentRound']
-        if project_info['isComplete']:
-            current_round += 1
+        project_key = project_info['key']
         ext = project_info['modelType']
-        prev_round = current_round - 1
+        current_round = project_info['currentRound']
+        prev_round = self._get_last_fed_model_round(current_round, project_info['isComplete'])
         federated_model_path = get_federated_path(model_path, ext, datasource_key, prev_round)
 
-        # Pull
-        print(f'>>> Waiting to download round ({prev_round}) federated model...')
-        project.pull_model(federated_model_path, round=prev_round)
+        yes_stats = len(self._check_stats(project_info['key'], datasource_key, prev_round))
+        yes_submission = len(self._check_submissions(project_info['key'], datasource_key, current_round))
+
+        if not yes_submission or not yes_stats:
+            # Pull
+            print(f'>>> ({project_key}-{datasource_key}) Waiting to download round ({prev_round}) federated model...')
+            project.pull_model(federated_model_path, round=prev_round)
 
         # Test
-        if test == None:
-            return
-        # if check_stats(current_round, datasource_key):
-        #     print(f'>>> Already uploaded round ({current_round}) stats.')
-        else:
-            print(f'>>> Running validation on round ({prev_round}) federated model...')
+        if not yes_stats:
+            print(f'>>> ({project_key}-{datasource_key}) Running validation on round ({prev_round}) federated model...')
             scores, num_samples = test(datasource_key, federated_model_path, project_info)
             print(scores)
-            print(f'>>> Uploading scores...')
+            print(f'>>> ({project_key}-{datasource_key}) Uploading scores...')
             project.report_stats(scores, num_samples, prev_round, datasource_key)
             print('Done.')
             print()
 
-        # Push
-        if not project_info['isComplete']:
+        # Train and push
+        if not yes_submission:
             new_model_path = get_trained_path(model_path, ext, datasource_key, current_round)
 
-            print('>>> Training weights on local model...')
+            print(f'>>> ({project_key}-{datasource_key}) Training weights on local model...')
             train(datasource_key, federated_model_path, new_model_path, project_info)
 
-            print(f'>>> Uploading round ({current_round}) trained model...')
+            print(f'>>> ({project_key}-{datasource_key}) Uploading round ({current_round}) trained model...')
             project.push_model(new_model_path, datasource_key)
             print('Done.')
             print()
 
-    def train_and_test_and_end_callback(self, datasource_key, project_info):
-        self.train_and_test_callback(datasource_key, project_info)
-        self.project_participants = list(filter(lambda x : x['project_key'] != project_info['key'], self.project_participants))
-
 
     # creates a new datasource in the api
     def attach_datasource(self, key, name=None, metadata=None):
@@ -402,6 +388,21 @@ class DynamoFL(_Base):
     def get_projects(self):
         return self._make_request('GET', '/projects', list=True)
 
+    def _check_submissions(self, project_key, datasource_key, round):
+        params = {
+            'owned': True,
+            'datasource': datasource_key,
+            'round': round
+        }
+        return self._make_request('GET', f'/projects/{project_key}/submissions', params, list=True)
+
+    def _check_stats(self, project_key, datasource_key, round):
+        params = {
+            'owned': True,
+            'datasource': datasource_key,
+            'round': round
+        }
+        return self._make_request('GET', f'/projects/{project_key}/stats', params, list=True)
 
 
 def get_trained_path(base, ext, ds, round):
